/*-------------------------------------------------------------------
 * nfpt: no-frills parallel transfer
 * a quick-and-dirty pseudo-ECP mode parallel cable transfer utility
 * Copyright (c) 2002,2003 K.H. Man <mkh@pl.jaring.my>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * ecp main file
 * Last revised: 2003-07-14
 *-------------------------------------------------------------------
 */

#include <stdio.h>
#include <stdlib.h>

#define ECP_C
#include "ecp.h"

#ifdef DEBUG
#define DEBUG_REG_STATUS
#define DEBUG_RX_PACKET
#define DEBUG_TX_PACKET
#endif

#ifdef ECP_TEST
#include <keys.h>
#endif

/*-------------------------------------------------------------------
 * debug routine(s)
 *-------------------------------------------------------------------
 */

#ifdef DEBUG_REG_STATUS
void print_status(void)
{
    int dcr, dsr, ecr;
    dcr = inportb(REGdcr);      /* find out status of major */
    dsr = inportb(REGdsr);      /* signal lines and states */
    ecr = inportb(REGecr);
    printf("dcr = [0x%02X]  dsr = [0x%02X]  "
           "ecr = [0x%02X]  stat = [0x%02X]\n",
           dcr, dsr, ecr, ECPstatus());
}
#define DS  print_status()
#else
#define DS
#endif

/*-------------------------------------------------------------------
 * crc lookup table
 * from util.c -- utility functions for gzip support
 * Copyright (C) 1992-1993 Jean-loup Gailly (GNU GPL)
 *-------------------------------------------------------------------
 */

ulg crc_32_tab[] = {
  0x00000000L, 0x77073096L, 0xee0e612cL, 0x990951baL, 0x076dc419L,
  0x706af48fL, 0xe963a535L, 0x9e6495a3L, 0x0edb8832L, 0x79dcb8a4L,
  0xe0d5e91eL, 0x97d2d988L, 0x09b64c2bL, 0x7eb17cbdL, 0xe7b82d07L,
  0x90bf1d91L, 0x1db71064L, 0x6ab020f2L, 0xf3b97148L, 0x84be41deL,
  0x1adad47dL, 0x6ddde4ebL, 0xf4d4b551L, 0x83d385c7L, 0x136c9856L,
  0x646ba8c0L, 0xfd62f97aL, 0x8a65c9ecL, 0x14015c4fL, 0x63066cd9L,
  0xfa0f3d63L, 0x8d080df5L, 0x3b6e20c8L, 0x4c69105eL, 0xd56041e4L,
  0xa2677172L, 0x3c03e4d1L, 0x4b04d447L, 0xd20d85fdL, 0xa50ab56bL,
  0x35b5a8faL, 0x42b2986cL, 0xdbbbc9d6L, 0xacbcf940L, 0x32d86ce3L,
  0x45df5c75L, 0xdcd60dcfL, 0xabd13d59L, 0x26d930acL, 0x51de003aL,
  0xc8d75180L, 0xbfd06116L, 0x21b4f4b5L, 0x56b3c423L, 0xcfba9599L,
  0xb8bda50fL, 0x2802b89eL, 0x5f058808L, 0xc60cd9b2L, 0xb10be924L,
  0x2f6f7c87L, 0x58684c11L, 0xc1611dabL, 0xb6662d3dL, 0x76dc4190L,
  0x01db7106L, 0x98d220bcL, 0xefd5102aL, 0x71b18589L, 0x06b6b51fL,
  0x9fbfe4a5L, 0xe8b8d433L, 0x7807c9a2L, 0x0f00f934L, 0x9609a88eL,
  0xe10e9818L, 0x7f6a0dbbL, 0x086d3d2dL, 0x91646c97L, 0xe6635c01L,
  0x6b6b51f4L, 0x1c6c6162L, 0x856530d8L, 0xf262004eL, 0x6c0695edL,
  0x1b01a57bL, 0x8208f4c1L, 0xf50fc457L, 0x65b0d9c6L, 0x12b7e950L,
  0x8bbeb8eaL, 0xfcb9887cL, 0x62dd1ddfL, 0x15da2d49L, 0x8cd37cf3L,
  0xfbd44c65L, 0x4db26158L, 0x3ab551ceL, 0xa3bc0074L, 0xd4bb30e2L,
  0x4adfa541L, 0x3dd895d7L, 0xa4d1c46dL, 0xd3d6f4fbL, 0x4369e96aL,
  0x346ed9fcL, 0xad678846L, 0xda60b8d0L, 0x44042d73L, 0x33031de5L,
  0xaa0a4c5fL, 0xdd0d7cc9L, 0x5005713cL, 0x270241aaL, 0xbe0b1010L,
  0xc90c2086L, 0x5768b525L, 0x206f85b3L, 0xb966d409L, 0xce61e49fL,
  0x5edef90eL, 0x29d9c998L, 0xb0d09822L, 0xc7d7a8b4L, 0x59b33d17L,
  0x2eb40d81L, 0xb7bd5c3bL, 0xc0ba6cadL, 0xedb88320L, 0x9abfb3b6L,
  0x03b6e20cL, 0x74b1d29aL, 0xead54739L, 0x9dd277afL, 0x04db2615L,
  0x73dc1683L, 0xe3630b12L, 0x94643b84L, 0x0d6d6a3eL, 0x7a6a5aa8L,
  0xe40ecf0bL, 0x9309ff9dL, 0x0a00ae27L, 0x7d079eb1L, 0xf00f9344L,
  0x8708a3d2L, 0x1e01f268L, 0x6906c2feL, 0xf762575dL, 0x806567cbL,
  0x196c3671L, 0x6e6b06e7L, 0xfed41b76L, 0x89d32be0L, 0x10da7a5aL,
  0x67dd4accL, 0xf9b9df6fL, 0x8ebeeff9L, 0x17b7be43L, 0x60b08ed5L,
  0xd6d6a3e8L, 0xa1d1937eL, 0x38d8c2c4L, 0x4fdff252L, 0xd1bb67f1L,
  0xa6bc5767L, 0x3fb506ddL, 0x48b2364bL, 0xd80d2bdaL, 0xaf0a1b4cL,
  0x36034af6L, 0x41047a60L, 0xdf60efc3L, 0xa867df55L, 0x316e8eefL,
  0x4669be79L, 0xcb61b38cL, 0xbc66831aL, 0x256fd2a0L, 0x5268e236L,
  0xcc0c7795L, 0xbb0b4703L, 0x220216b9L, 0x5505262fL, 0xc5ba3bbeL,
  0xb2bd0b28L, 0x2bb45a92L, 0x5cb36a04L, 0xc2d7ffa7L, 0xb5d0cf31L,
  0x2cd99e8bL, 0x5bdeae1dL, 0x9b64c2b0L, 0xec63f226L, 0x756aa39cL,
  0x026d930aL, 0x9c0906a9L, 0xeb0e363fL, 0x72076785L, 0x05005713L,
  0x95bf4a82L, 0xe2b87a14L, 0x7bb12baeL, 0x0cb61b38L, 0x92d28e9bL,
  0xe5d5be0dL, 0x7cdcefb7L, 0x0bdbdf21L, 0x86d3d2d4L, 0xf1d4e242L,
  0x68ddb3f8L, 0x1fda836eL, 0x81be16cdL, 0xf6b9265bL, 0x6fb077e1L,
  0x18b74777L, 0x88085ae6L, 0xff0f6a70L, 0x66063bcaL, 0x11010b5cL,
  0x8f659effL, 0xf862ae69L, 0x616bffd3L, 0x166ccf45L, 0xa00ae278L,
  0xd70dd2eeL, 0x4e048354L, 0x3903b3c2L, 0xa7672661L, 0xd06016f7L,
  0x4969474dL, 0x3e6e77dbL, 0xaed16a4aL, 0xd9d65adcL, 0x40df0b66L,
  0x37d83bf0L, 0xa9bcae53L, 0xdebb9ec5L, 0x47b2cf7fL, 0x30b5ffe9L,
  0xbdbdf21cL, 0xcabac28aL, 0x53b39330L, 0x24b4a3a6L, 0xbad03605L,
  0xcdd70693L, 0x54de5729L, 0x23d967bfL, 0xb3667a2eL, 0xc4614ab8L,
  0x5d681b02L, 0x2a6f2b94L, 0xb40bbe37L, 0xc30c8ea1L, 0x5a05df1bL,
  0x2d02ef8dL
};

/*-------------------------------------------------------------------
 * calculates the crc of the given array of BYTES
 * modified from util.c -- utility functions for gzip support
 * Copyright (C) 1992-1993 Jean-loup Gailly (GNU GPL)
 * ulg is assumed to be 32 bits; no longer does running crc
 *-------------------------------------------------------------------
 */

ulg crc(BYTE *buf, int size)
{
    register ulg crc = (ulg)0xffffffffL;

    if (size)
    {
        do
        {
            crc = crc_32_tab[((int)crc ^ (*buf++)) & 0xFF]
                  ^ (crc >> 8);

        } while (--size);
    }
    return crc;
}

/*-------------------------------------------------------------------
 * sets the mode of the LPT port, usually between PS2 and ECP mode
 *-------------------------------------------------------------------
 */

void ECPmode(int MODE)
{
    register int ecr;

    ecr = inportb(REGecr);
    ecr = (ecr & 0x1F) | (MODE << 5);   /* ecr[7..5] is mode bits */
    outportb(REGecr, ecr);
}

/*-------------------------------------------------------------------
 * sets the direction of data transfer, best done in PS2 mode
 *-------------------------------------------------------------------
 */

void ECPdirection(int DIRECTION)
{
    register int dcr;

    dcr = inportb(REGdcr);
    dcr = (dcr & 0xDF) | DIRECTION;     /* dcr[5] is direction bit */
    outportb(REGdcr, dcr);
}

/*-------------------------------------------------------------------
 * tests I/O register addrs for an ECP port, sets global variables
 * returns ECP_SUCCESS or ECP_ERR_*
 *-------------------------------------------------------------------
 */

int ECPtest(void)
{
    int i, j, result;
    int dcr, ecr, cnfgA, cnfgB;

    result = ECP_SUCCESS;           /* result value holder */

    /* ----------------------------------------------------
     * first, we identify the ECP port address
     */
    for (j = 0; j < 2; j++)         /* 2-pass logic */
    {
        for (i = 0; i < ECP_LIST; i++)
        {
            REGbase = ECPaddrs[i];  /* check for presence of ecr */
            ecr = inportb(REGecr);

            if ((ecr & 0x03) == 1)  /* full=0,empty=1 */
            {                       /* as per MS doc */
                outportb(REGecr, 0x34);
                if (inportb(REGecr) == 0x35)
                {
                    break;          /* found, index in i */
                }
            }
        }
        if (i == ECP_LIST)      /* if not found, the port */
        {                       /* could be 'hanging' */

            for (i = 0; i < ECP_LIST; i++)
            {
                REGbase = ECPaddrs[i];
                ecr = inportb(REGecr);

                if (ecr != 0xFF)/* if it's not 0xFF, then */
                {               /* it is probably a port */
                    ECPreset(); /* try to reset to PS2 mode */
                }
            }                   /* now we can try again... */
        }
        else
        {
            break;              /* exit immediately if found */
        }
    }
    if (i == ECP_LIST)
    {
        return ECP_ERR_NOT_FOUND;   /* this is fatal (first one) */
    }

    /* ----------------------------------------------------
     * next, we look at its configuration
     */
    ECPmode(MODEconfig);        /* look at some capabilities */
    cnfgA = inportb(REGcnfgA);
    cnfgB = inportb(REGcnfgB);
    pwordsz = (cnfgA & 0x70) >> 4;  /* port word size */

    if (pwordsz != 1)               /* check port word size */
    {
        result |= ECP_ERR_PWORD8;   /* non-fatal */
    }

    if (((cnfgB & 0x40) >> 6) == 0) /* interrupt conflict problem */
    {
        result |= ECP_ERR_CONFLICT; /* non-fatal */
    }

    /* ----------------------------------------------------
     * then, check the IRQ and DMA settings if possible
     */
    irq = IRQlookup[(cnfgB & 0x38) >> 3];
    dma = DMAlookup[(cnfgB & 0x03)];

    if (irq == 0)   /* set defaults if port does not say */
    {               /* in MS doc, this is set by jumpers */

        irq = IRQdefault;
        result |= ECP_ERR_IRQ;
    }
    if ((dma == 0) || (dma == 4))
    {               /* same with DMA reporting */

        dma = DMAdefault;
        result |= ECP_ERR_DMA;
    }

    /* ----------------------------------------------------
     * test direction bit
     */
    ECPmode(MODEps2);

    dcr = inportb(REGdcr);          /* see if direction bit */
    outportb(REGdcr, dcr | 0x20);   /* is behaving */
    dcr = inportb(REGdcr);
    outportb(REGdcr, dcr & 0xDF);

    if ((dcr & 0x20) != 0x20)       /* can switch W->R->W? */
    {
        result |= ECP_ERR_DIRECTION;
    }

    /* ----------------------------------------------------
     * find out size of FIFO buffer
     */
    ECPmode(MODEps2);               /* algorithm from MS */
    ECPdirection(DIRECTIONwrite);   /* document */
    ECPmode(MODEtest);

    i = 0;                          /* write to FIFO until */
    do {                            /* it is full, warning */
        outportb(REGecpDFifo, 0);   /* this uses 8 bit words */
        i++;
        ecr = inportb(REGecr);
    } while ((ecr & 0x02) != 0x02);

    ECPmode(MODEps2);
    FIFO = i;
    
    if (FIFO < 8)       /* make sure got some form of FIFO */
    {
        result |= ECP_ERR_FIFO;
    }

    /* ---------------------------------------------------- */
    return result;      /* config done, globals set */
}

/*-------------------------------------------------------------------
 * initialize ECP port, we assume configuration (test) went well
 * we stay in PS2 mode, since we are listening on the signal lines
 *-------------------------------------------------------------------
 */

void ECPinit(void)
{
    int dcr;

    ECPmode(MODEps2);           /* set direction */
    ECPdirection(DIRECTIONread);/* defaults to read (listen) */

    dcr = inportb(REGdcr);
    dcr = (dcr & MASKdcr) | SETBnInit;
                                /* deassert all signals */
    outportb(REGdcr, dcr);      /* from MS document */
                                /* redundant calls here... */
    ECPlisten();
}

/*-------------------------------------------------------------------
 * exit ECP port mode at the end of it all (maybe reset is better?)
 *-------------------------------------------------------------------
 */

void ECPexit(void)
{
    ECPmode(MODEps2);           /* end of ecp mode */
}

/*-------------------------------------------------------------------
 * reset ECP port after a botched transmission, clears the FIFO
 *-------------------------------------------------------------------
 */

void ECPreset(void)
{
    int dcr;

    dcr = inportb(REGdcr);
    dcr = dcr | SETBstrobe;     /* nStrobe is driven low */
    outportb(REGdcr, dcr);      /* from MS document */

    ECPmode(MODEps2);           /* reset to PS2 mode */

    ECPinit();                  /* set it back to initial state */
}

/*-------------------------------------------------------------------
 * returns status bits of dcr and ecr for easy testing using STAT*
 *-------------------------------------------------------------------
 */

int ECPstatus(void)
{
    int result, dsr, ecr;
    
    dsr = inportb(REGdsr);      /* add dcr signal line bits */
    result = dsr & 0xF8;
    
    ecr = inportb(REGecr);      /* add ecr FIFO bits (full, empty) */
    result |= ecr & 0x03;

    return result;
}

/*-------------------------------------------------------------------
 * custom handshake: receiver listen waits for an incoming
 *
 *                              (nAck)  (busy)  (PError)  <-- Read
 *                              nStrobe nAutoFd nInit     <-- Write
 * (Tx) BASE        listen          0       1       0
 *-------------------------------------------------------------------
 */

void ECPlisten(void)
{
    register int dcr;
    
    ECPmode(MODEps2);               /* set direction */
    ECPdirection(DIRECTIONread);    /* defaults to read (listen) */

    dcr = inportb(REGdcr);
    dcr = dcr & (MASKstrobe & MASKautoFd & MASKnInit)
              | (SETBstrobe);
    outportb(REGdcr, dcr);
}

/*-------------------------------------------------------------------
 * custom handshake: sender tells receiver to get set, base state
 *
 *                              (nAck)  (busy)  (PError)  <-- Read
 *                              nStrobe nAutoFd nInit     <-- Write
 * (--) SENDS       getready        0       1       1
 *-------------------------------------------------------------------
 */

void ECPgetready(void)
{
    register int dcr;
    
    dcr = inportb(REGdcr);
    dcr = dcr & (MASKstrobe & MASKautoFd & MASKnInit)
              | (SETBstrobe | SETBnInit);
    outportb(REGdcr, dcr);
}

/*-------------------------------------------------------------------
 * custom handshake: receiver detects an incoming request
 *
 *                              (nAck)  (busy)  (PError)  <-- Read
 *                              nStrobe nAutoFd nInit     <-- Write
 * (Rx) WAITS FOR   incoming        0       1       1
 *-------------------------------------------------------------------
 */

int ECPincoming(void)
{
    register int es;

    es = ECPstatus();
    if (((es & STATnAck) == 0) &&
        ((es & STATnBusy) == 0) &&              /* inverted */
        ((es & STATPError) == STATPError))
    {
        return 1;
    }
    return 0;
}

/*-------------------------------------------------------------------
 * custom handshake: sender detects receiver acknowledgement of ready
 *
 *                              (nAck)  (busy)  (PError)  <-- Read
 *                              nStrobe nAutoFd nInit     <-- Write
 * (Tx) WAITS FOR   ack             1       1       0
 *-------------------------------------------------------------------
 */

int ECPack(void)
{
    register int es;

    es = ECPstatus();
    if (((es & STATnAck) == STATnAck) &&
        ((es & STATnBusy) == 0) &&              /* inverted */
        ((es & STATPError) == 0))
    {
        return 1;
    }
    return 0;
}

/*-------------------------------------------------------------------
 * custom handshake: sender detects receiver back in listen mode
 *
 *                              (nAck)  (busy)  (PError)  <-- Read
 *                              nStrobe nAutoFd nInit     <-- Write
 * (Tx) WAITS FOR   complete        0       1       0
 *-------------------------------------------------------------------
 */

int ECPcomplete(void)
{
    register int es;

    es = ECPstatus();
    if (((es & STATnAck) == 0) &&
        ((es & STATnBusy) == 0) &&              /* inverted */
        ((es & STATPError) == 0))
    {
        return 1;
    }
    return 0;
}

/*-------------------------------------------------------------------
 * custom handshake: receiver reports it is ready for reception
 *
 *                              (nAck)  (busy)  (PError)  <-- Read
 *                              nStrobe nAutoFd nInit     <-- Write
 * (Rx) SENDS       ready           1       1       0
 *-------------------------------------------------------------------
 */

void ECPready(void)
{
    register int dcr;
    
    dcr = inportb(REGdcr);
    dcr = dcr & (MASKstrobe & MASKautoFd & MASKnInit);
    outportb(REGdcr, dcr);

    ECPdirection(DIRECTIONread);    /* set direction (redundant) */
    ECPmode(MODEecp);               /* ECP mode, ready to receive */
}

/*-------------------------------------------------------------------
 * custom handshake: sender sets the signal lines to start transfer
 *
 *                              (nAck)  (busy)  (PError)  <-- Read
 *                              nStrobe nAutoFd nInit     <-- Write
 * (Tx) SETS        start           1       1       1
 *-------------------------------------------------------------------
 */

void ECPstart(void)
{
    register int dcr;
    
    dcr = inportb(REGdcr);
    dcr = dcr & (MASKstrobe & MASKautoFd & MASKnInit)
              | (SETBnInit);
    outportb(REGdcr, dcr);

    ECPdirection(DIRECTIONwrite);   /* set direction */
    ECPmode(MODEecp);               /* ECP mode, ready to transmit */
}

/*-------------------------------------------------------------------
 * sets the crc value of the header segment, must do before checksum
 *-------------------------------------------------------------------
 */

void ECPcrc(struct HEADER *hd, BYTE *buf, int buflen)
{
    hd->crc = crc(buf, buflen);
}

/*-------------------------------------------------------------------
 * calculate a checksum of header, does NOT set it in header field
 * this must be done after crc calculation
 *-------------------------------------------------------------------
 */

BYTE ECPchecksum(struct HEADER *hd)
{
    register int sum;
    
    sum = hd->id + hd->cmd
        + hd->slicesize + hd->slices
        + hd->info1 + hd->info2 + hd->info3

        + (hd->crc & 0xFF)                  /* crc */
        + ((hd->crc >>  8) & 0xFF)
        + ((hd->crc >> 16) & 0xFF)
        + ((hd->crc >> 24) & 0xFF)
        + (hd->count & 0xFF)

        + ((hd->count >>  8) & 0xFF)        /* count */
        + ((hd->count >> 16) & 0xFF)
        + ((hd->count >> 24) & 0xFF);
    
    return sum;
}

/*-------------------------------------------------------------------
 * write header data byte-by-byte to a BYTE buffer
 *-------------------------------------------------------------------
 */

void ECPwriteHeader(struct HEADER *hd, BYTE *buf)
{
    *(buf  ) = hd->id;
    *(buf+1) = hd->cmd;
    *(buf+2) = hd->chksum;
    *(buf+3) = hd->slicesize;
    *(buf+4) = hd->slices;
    *(buf+5) = hd->info1;
    *(buf+6) = hd->info2;
    *(buf+7) = hd->info3;

    *(buf+ 8) = hd->crc & 0xFF;             /* crc */
    *(buf+ 9) = (hd->crc >>  8) & 0xFF;
    *(buf+10) = (hd->crc >> 16) & 0xFF;
    *(buf+11) = (hd->crc >> 24) & 0xFF;
    
    *(buf+12) = hd->count & 0xFF;           /* count */
    *(buf+13) = (hd->count >>  8) & 0xFF;
    *(buf+14) = (hd->count >> 16) & 0xFF;
    *(buf+15) = (hd->count >> 24) & 0xFF;
}

/*-------------------------------------------------------------------
 * read header data byte-by-byte from a BYTE buffer
 *-------------------------------------------------------------------
 */

void ECPreadHeader(struct HEADER *hd, BYTE *buf)
{
    hd->id          = *(buf);
    hd->cmd         = *(buf+1);
    hd->chksum      = *(buf+2);
    hd->slicesize   = *(buf+3);
    hd->slices      = *(buf+4);
    hd->info1       = *(buf+5);
    hd->info2       = *(buf+6);
    hd->info3       = *(buf+7);
    
    hd->crc =  *(buf+ 8)                    /* crc */
            | (*(buf+ 9) <<  8)
            | (*(buf+10) << 16)
            | (*(buf+11) << 24);

    hd->count =  *(buf+12)                  /* count */
              | (*(buf+13) <<  8)
              | (*(buf+14) << 16)
              | (*(buf+15) << 24);
}

/*-------------------------------------------------------------------
 * prepare a command packet header, no checksum or crc done yet
 *-------------------------------------------------------------------
 */

void ECPcmdPacket(struct HEADER *hd, int COMMAND)
{
    hd->id = HEADER_ID;     /* ID */
    hd->cmd = COMMAND;      /* command */
    hd->chksum = 0;         /* no checksum (set last) */

    hd->slicesize = FIFO;   /* zero length data section */
    hd->slices = 0;

    hd->info1 = 0;          /* no info */
    hd->info2 = 0;
    hd->info3 = 0;

    hd->count = ++pktcount; /* sequence this packet */
    hd->crc = 0;            /* no crc (set later) */
}

/*-------------------------------------------------------------------
 * build a null packet (for testing, etc.), payload optional
 *-------------------------------------------------------------------
 */

void ECPnullPacket(struct HEADER *hd, int size)
{
    ECPcmdPacket(hd, CMDnull);
    hd->slices = (size + FIFO - 1) / FIFO;      /* round up */
}

/*-------------------------------------------------------------------
 * build a hello packet (initiates connection)
 *-------------------------------------------------------------------
 */

void ECPhelloPacket(struct HEADER *hd)
{
    ECPcmdPacket(hd, CMDhello);
}

/*-------------------------------------------------------------------
 * build a goodbye packet (ends connection)
 *-------------------------------------------------------------------
 */

void ECPgoodbyePacket(struct HEADER *hd)
{
    ECPcmdPacket(hd, CMDgoodbye);
}

/*-------------------------------------------------------------------
 * build a terminate packet (resets something, e.g. a file transfer)
 *-------------------------------------------------------------------
 */

void ECPterminatePacket(struct HEADER *hd)
{
    ECPcmdPacket(hd, CMDterminate);
}

/*-------------------------------------------------------------------
 * sends a packet to the remote machine, always sets crc & chksum
 * buflen should be equals to slicesize*slices
 *-------------------------------------------------------------------
 */

int ECPsendPacket(struct HEADER *hd, BYTE *buf, int buflen)
{
    int ticks, result;
    int ecp_timeout = ECP_TX_TIMEOUT;
    BYTE hdbuf[HEADER_SIZE];
    
    /* ----------------------------------------------------
     * set size of data to receive
     */
    if ((buflen > 0) && (hd->slices == 0))
    {
	hd->slices = (buflen + FIFO - 1) / FIFO;      /* round up */
	buflen = hd->slices * FIFO;
    }

    /* ----------------------------------------------------
     * packet preparation
     */
    if (hd->slices)
    {
	ECPcrc(hd, buf, buflen);        /* sets the crc */
    }
    hd->chksum = ECPchecksum(hd);       /* sets checksum */
    ECPwriteHeader(hd, hdbuf);          /* converts header to bytes */

    /* ----------------------------------------------------
     * handshake -- request, then wait for acknowledgement
     */
#ifdef DEBUG
    printf("Write handshake (listen)\n"); DS;
#endif
    ECPgetready();                      /* request to send */
#ifdef DEBUG
    printf("Write handshake (getready)\n"); DS;
#endif

    ticks = rawclock() + ecp_timeout;
    do
    {
    } while ((!ECPack()) &&             /* wait for acknowledge */
             (rawclock() < ticks));

    if (rawclock() >= ticks)
    {
#ifdef DEBUG
        printf("Acknowledge error\n"); DS;
#endif
        ECPreset();                     /* error */
        return ECP_ERR_NOACK;
    }

    ECPstart();                         /* set ECP initial state */
#ifdef DEBUG
    printf("Write handshake (start)\n"); DS;
#endif

    /* ----------------------------------------------------
     * transmission
     */
    result = ECPwriteBlock(hdbuf, HEADER_SIZE);
                                        /* sends the header */
    if (result != ECP_SUCCESS)
    {
#ifdef DEBUG
        printf("Error writing header block\n");
#endif
        ECPreset();                     /* error */
        return result;
    }
    if (hd->slices > 0)
    {
        result = ECPwriteBlock(buf, buflen);    /* sends the data */
        if (result != ECP_SUCCESS)
        {
#ifdef DEBUG
            printf("Error writing data block\n");
#endif
            ECPreset();                 /* error */
            return result;
        }
    }

    /* ----------------------------------------------------
     * wait for completion
     */
    ticks = rawclock() + ecp_timeout;
    do
    {
    } while ((!ECPcomplete()) &&        /* wait for completion */
             (rawclock() < ticks));

    if (rawclock() >= ticks)
    {
#ifdef DEBUG
        printf("Error waiting for completion\n");
#endif
        ECPreset();                     /* error */
        return ECP_ERR_NOCOMPLETE;
    }
#ifdef DEBUG
    printf("Success sending block\n"); DS;
#endif

    ECPlisten();                        /* packet sent, listen now */
#ifdef DEBUG
    DS; printf("------------------------------------\n");
#endif
    return ECP_SUCCESS;                 /* all OK */
}

/*-------------------------------------------------------------------
 * receives a packet from the remote machine, tests crc & chksum
 *-------------------------------------------------------------------
 */

int ECPreceivePacket(struct HEADER *hd, BYTE *buf, int *buflen)
{
    int j;
    int ticks, result, size;
    int ecp_timeout = ECP_RX_TIMEOUT;
    BYTE chksum;
    ulg crc_value;
    BYTE hdbuf[HEADER_SIZE];

    /* ----------------------------------------------------
     * handshake sequence, checks or confirms incoming
     */
#ifdef DEBUG
    printf("Read handshake (listen)\n"); DS;
#endif
    ticks = rawclock() + ecp_timeout;
    do
    {
    } while ((!ECPincoming()) &&        /* wait for incoming */
             (rawclock() < ticks));

    if (rawclock() >= ticks)
    {
#ifdef DEBUG
        printf("Incoming error\n"); DS;
#endif
        ECPreset();                     /* error */
        return ECP_ERR_NOINCOMING;
    }
#ifdef DEBUG
    printf("Read handshake (incoming)\n"); DS;
#endif
    ECPready();                         /* ready to receive state */
#ifdef DEBUG
    printf("Read handshake (ready)\n"); DS;
#endif

    /* ----------------------------------------------------
     * receive header & tests integrity using checksum
     */
    result = ECPreadBlock(hdbuf, HEADER_SIZE);
                                        /* receives the header */
    if (result != ECP_SUCCESS)
    {
#ifdef DEBUG
        printf("Error reading header block\n");
#endif
        ECPreset();                     /* error */
        return result;
    }

    int garbage = 0;
    while (hdbuf[0] != HEADER_ID)	/* starting glitch dunno why it */
    {					/* sometimes appears? */
	garbage++;
	if (garbage == ECP_RETRIES)
	{
#ifdef DEBUG
	    printf("Checksum read error:\n");
	    for (j = 0; j < FIFO; j++)
	    {
		printf("IN[0x%2X]  ", hdbuf[j]);
	    }
#endif
	    ECPreset();                     /* error */
	    return ECP_ERR_CHECKSUM;
	}
        for (j = 0; j < HEADER_SIZE-1; j++)
        {
            hdbuf[j] = hdbuf[j + 1];
        }
        hdbuf[j] = inportb(REGecpDFifo);
    }

    ECPreadHeader(hd, hdbuf);           /* converts bytes to struct */

    chksum = ECPchecksum(hd);           /* checksum test */
    if (chksum != hd->chksum)
    {
#ifdef DEBUG
        printf("Checksum error:\n");
        for (j = 0; j < FIFO; j++)
        {
            printf("IN[0x%2X]  ", hdbuf[j]);
        }
#endif
        ECPreset();                     /* error */
        return ECP_ERR_CHECKSUM;
    }
    size = hd->slices * hd->slicesize;  /* size of block */
    *buflen = size;

#ifdef DEBUG
    printf("Checksum good, data to read = %dB\n", size); DS;
#endif
    /* ----------------------------------------------------
     * receives data & tests integrity using crc
     */
    if (size > 0)
    {
        result = ECPreadBlock(buf, size);   /* receives the data */

        if (result != ECP_SUCCESS)
        {
#ifdef DEBUG
            printf("Error reading data block\n"); DS;
#endif
            ECPreset();                 /* error */
            return result;
        }

        crc_value = crc(buf, *buflen);  /* crc test */
        if (crc_value != hd->crc)
        {
            ECPreset();                 /* error */
            return ECP_ERR_CRC;
        }
    }

#ifdef DEBUG
    printf("Success receiving block\n"); DS;
#endif
    ECPlisten();                        /* packet rxd, listen now */
#ifdef DEBUG
    DS; printf("------------------------------------\n");
#endif
    return ECP_SUCCESS;                 /* all OK */
}

/*-------------------------------------------------------------------
 * sends bytes less than FIFO size (!! must have size < FIFO)
 *-------------------------------------------------------------------
 */

int ECPwriteBytes(BYTE *buf, int size)
{
    int ticks, ecr;
    int ecp_timeout = ECP_IO_TIMEOUT;

    ticks = rawclock() + ecp_timeout;
            
    do                                  /* wait for empty */
    {
        ecr = inportb(REGecr);

    } while (((ecr & STATempty) != STATempty) &&
             (rawclock() < ticks));

    if (rawclock() >= ticks)            /* timeout occured, error */
    {
        return ECP_ERR_TIMEOUT;
    }
    outportsb(REGecpDFifo, buf, size);  /* send data (byte) */

    /* ---------------------------------------------------- */
    do                              /* wait for empty before exiting */
    {
        ecr = inportb(REGecr);

    } while (((ecr & STATempty) != STATempty) &&
             (rawclock() < ticks));

    if (rawclock() >= ticks)        /* timeout occured, error */
    {
        return ECP_ERR_TIMEOUT;
    }
    return ECP_SUCCESS;
}

/*-------------------------------------------------------------------
 * sends a block of data, using FIFO-sized slices if possible
 *-------------------------------------------------------------------
 */

int ECPwriteBlock(BYTE *buf, int size)
{
    int i, j, slices, residue, ecr;
    int ecp_timeout = ECP_IO_TIMEOUT;
    long ticks;

    if (size < FIFO)                /* for short runs */
    {
        return ECPwriteBytes(buf, size);
    }
    slices = size / FIFO;           /* calc number of slices */
    residue = size % FIFO;
    
    for (i = 0; i < slices; i++)    /* per-FIFO transmit */
    {
        ticks = rawclock() + ecp_timeout;

        do                          /* wait for empty */
        {
            ecr = inportb(REGecr);

        } while (((ecr & STATempty) != STATempty) &&
                 (rawclock() < ticks));

        if (rawclock() >= ticks)    /* timeout occured, error */
        {
            return ECP_ERR_TIMEOUT;
        }
#ifdef PORT_IO_CALL
        outportsb(REGecpDFifo, buf, FIFO);  /* send data block */
        buf += FIFO;                        /* update pointer */
#endif
#ifdef PORT_IO_LOOP
        for (j = 0; j < FIFO; j++)
        {
            outportb(REGecpDFifo, *buf);
#ifdef DEBUG
	    printf("OUT[0x%2X] ", *buf);
#endif
            buf++;
        }
#endif
    }
    if (residue)                    /* now for the balance */
    {
        return ECPwriteBytes(buf, residue);
    }
    /* ---------------------------------------------------- */
    do                              /* wait for empty before exiting */
    {
        ecr = inportb(REGecr);

    } while (((ecr & STATempty) != STATempty) &&
             (rawclock() < ticks));

    if (rawclock() >= ticks)        /* timeout occured, error */
    {
        return ECP_ERR_TIMEOUT;
    }
    return ECP_SUCCESS;
}

/*-------------------------------------------------------------------
 * receives bytes less than FIFO size (!! must have size < FIFO)
 *-------------------------------------------------------------------
 */

int ECPreadBytes(BYTE *buf, int size)
{
    BYTE b;
    int i, ticks, ecr;
    int ecp_timeout = ECP_IO_TIMEOUT;

    for (i = 0; i < size; i++)
    {
        ticks = rawclock() + ecp_timeout;
            
        do                              /* wait for at least one */
        {
            ecr = inportb(REGecr);

        } while (((ecr & STATempty) == STATempty) &&
                 (rawclock() < ticks));

        if (rawclock() >= ticks)        /* timeout occured, error */
        {
            return ECP_ERR_TIMEOUT;
        }
        *buf = inportb(REGecpDFifo);    /* receive data (byte) */

        buf++;
     }
     return ECP_SUCCESS;
}

/*-------------------------------------------------------------------
 * receives a block of data, using FIFO-sized slices if possible
 *-------------------------------------------------------------------
 */

int ECPreadBlock(BYTE *buf, int size)
{
    int i, j, slices, residue, ecr;
    int ecp_timeout = ECP_IO_TIMEOUT;
    long ticks;

    if (size < FIFO)                /* for short runs */
    {
        return ECPreadBytes(buf, size);
    }
    slices = size / FIFO;           /* calc number of slices */
    residue = size % FIFO;

    for (i = 0; i < slices; i++)    /* per-FIFO receive */
    {
        ticks = rawclock() + ecp_timeout;

        do                          /* wait for full */
        {
            ecr = inportb(REGecr);

        } while (((ecr & STATfull) != STATfull) &&
                 (rawclock() < ticks));

        if (rawclock() >= ticks)    /* timeout occured, error */
        {
            return ECP_ERR_TIMEOUT;
        }
#ifdef PORT_IO_CALL
        inportsb(REGecpDFifo, buf, FIFO);   /* receive data block */
        buf += FIFO;                        /* update pointer */
#endif
#ifdef PORT_IO_LOOP
        for (j = 0; j < FIFO; j++)
        {
            *buf = inportb(REGecpDFifo);
#ifdef DEBUG
	     printf("IN[0x%2X]  ", *buf);
#endif
            buf++;
        }
#endif
    }
    if (residue)                    /* now for the balance */
    {
        return ECPreadBytes(buf, residue);
    }
    return ECP_SUCCESS;
}

/*-------------------------------------------------------------------
 * make a block of pesudo-random byte data
 *-------------------------------------------------------------------
 */

BYTE block_make(BYTE *buf, int buflen)
{
    int i, seed;
    
    seed = rawclock() & 0xFF;       /* set a pseudo sequence */
    srand(seed);
    
    for (i = 0; i < buflen; i++)    /* make the pseudo sequence */
    {
        *buf = rand() % 0x100;
        buf++;
    }
    return seed;                    /* return seed for header */
}

/*-------------------------------------------------------------------
 * compare a block of pseudo-random byte data
 *-------------------------------------------------------------------
 */

int block_compare(BYTE *buf, int buflen, BYTE seed)
{
    int i, errors;
    
    errors = 0;                     /* start with zero errors */
    
    srand(seed);                    /* set the same sequence */
    
    for (i = 0; i < buflen; i++)    /* compare the pseudo sequence */
    {
        if ((rand() % 0x100) != *buf)
        {
            errors++;
        }
        buf++;
    }
    return errors;                  /* return number of errors */
}

/*-------------------------------------------------------------------
 * code for testing the connection
 *-------------------------------------------------------------------
 */

#ifdef ECP_TEST

BYTE data_tx[MAX_PACKET];   /* transmitted sequence */
struct HEADER head;         /* header for this test */

/*-------------------------------------------------------------------
 * transmit test routine
 *-------------------------------------------------------------------
 */

void transmit_test(void)
{
    int i, blocklen, ticks, rate, result;

    blocklen = BLK_SIZE * FIFO;         /* block size */
    ticks = rawclock();                 /* start of test */

    for (i = 0; i < BLK_COUNT; i++)     /* run a while, fixed blks */
    {
        ECPcmdPacket(&head, CMDnull);   /* build packet */
        head.slices = BLK_SIZE;
        head.info1 = block_make(data_tx, blocklen);

        result = ECPsendPacket(&head, data_tx, blocklen);
                                        /* transmit */

        switch (result)
        {
	    case ECP_SUCCESS:
		break;
            case ECP_ERR_TIMEOUT:
                printf("Timeout error at block %d\n", i);
                return;
            case ECP_ERR_NOACK:
                printf("No acknowledge when starting block %d\n", i);
                return;
            case ECP_ERR_NOCOMPLETE:
                printf("No complete when ending block %d\n", i);
                return;
	    default:
                printf("Unknown error handling block %d\n", i);
                return;
        }
    }
    ticks = rawclock() - ticks;         /* get ticks elapsed */
    if (ticks == 0)                     /* avoid division by 0 */
    {
        ticks++;
    }
    rate = blocklen * i / ticks         /* calculates K/s unit */
           * 182 / 10 / 1024;
    
    printf("\nBlocks of %d bytes transmitted in %d ticks: %d\n"
           "Transmission data rate: %d\n\n",
           blocklen, ticks, i, rate);   /* report */
}

/*-------------------------------------------------------------------
 * receive test routine
 *-------------------------------------------------------------------
 */

void receive_test(void)
{
    int i, blocklen, ticks, rate, result, errors;
    
    ticks = rawclock();                 /* start tick value */
    
    for (i = 0; i < BLK_COUNT; i++)     /* run a while, fixed blks */
    {
        result = ECPreceivePacket(&head, data_tx, &blocklen);
                                        /* read a packet */
        if (result != ECP_SUCCESS)
        {
            switch (result)             /* handle rxd errors */
            {
	    case ECP_SUCCESS:
		break;
            case ECP_ERR_NOINCOMING:
                printf("Incoming error at block %d\n", i);
                break;
            case ECP_ERR_TIMEOUT:
                printf("Timeout error at block %d\n", i);
                break;
            case ECP_ERR_CHECKSUM:
                printf("Checksum error (header) at block %d\n", i);
                break;
            case ECP_ERR_CRC:
                printf("CRC error (data) at block %d\n", i);
                break;
	    default:
                printf("Unknown error handling block %d\n", i);
                return;
            }
            break;                      /* terminate rxd */
        }
        else                            /* check data correctness */
        {
            errors = block_compare(data_tx, blocklen, head.info1);
            if (errors > 0)
            {
                printf("%d data error(s) at block %d\n", errors, i);
            }
        }
    }
    ticks = rawclock() - ticks;         /* get ticks elapsed */
    if (ticks == 0)                     /* avoid division by 0 */
    {
        ticks++;
    }
    rate = blocklen * i / ticks         /* calculates K/s unit */
           * 182 / 10 / 1024;
    
    printf("\nBlocks of %d bytes received in %d ticks: %d\n"
           "Transmission data rate: %d\n\n",
           blocklen, ticks, i, rate);   /* report */
}

/*-------------------------------------------------------------------
 * dummy transmit test routine
 *-------------------------------------------------------------------
 */

void dummy_test(void)
{
    int i, blocklen, ticks, rate, result;

    blocklen = BLK_SIZE * FIFO;         /* block size */
    ticks = rawclock();                 /* start of test */

    for (i = 0; i < BLK_COUNT; i++)     /* run a while, fixed blks */
    {
        ECPcmdPacket(&head, CMDnull);   /* build packet */
        head.slices = BLK_SIZE;
        head.info1 = block_make(data_tx, blocklen);

        result = ECPsendDummy(&head, data_tx, blocklen);
                                        /* transmit */
    }
    ticks = rawclock() - ticks;         /* get ticks elapsed */
    if (ticks == 0)                     /* avoid division by 0 */
    {
        ticks++;
    }
    rate = blocklen * i / ticks         /* calculates K/s unit */
           * 182 / 10 / 1024;
    
    printf("\nBlocks of %d bytes transmitted in %d ticks: %d\n"
           "Transmission data rate: %d\n\n",
           blocklen, ticks, i, rate);   /* report */
}

/*-------------------------------------------------------------------
 * dummy send tester
 *-------------------------------------------------------------------
 */

int ECPsendDummy(struct HEADER *hd, BYTE *buf, int buflen)
{
    int ticks, result;
    BYTE hdbuf[HEADER_SIZE];

    ECPcrc(hd, buf, buflen);            /* sets the crc */
    hd->chksum = ECPchecksum(hd);       /* sets checksum */
    ECPwriteHeader(hd, hdbuf);          /* converts header to bytes */

    return ECP_SUCCESS;                 /* all OK */
}

/*-------------------------------------------------------------------
 * handshake signal tester
 *-------------------------------------------------------------------
 */

void signal_test(void)
{
    int kp, dcr, dsr;
    
    printf("Select one write option [F1] to [F5] and "
           "read the result on the\n"
           "remote computer using [F6]...\n");

    while (1)                   /* main loop */
    {
        printf(
    "[F1]  STAND DOWN all lines, seen as DEASSERTED on remote machine\n"
    "[F2]  ASSERT   [0] nStrobe (pin 1)    --> ASSERT [0] nAck (pin 10)\n"
    "[F3]  DEASSERT [1] nAutoFd (pin 14)   --> ASSERT [1] busy (pin 11)\n"
    "[F4]  DEASSERT [1] nInit (pin 16)     --> ASSERT [1] PError (pin 12)\n"
    "[F5]  ASSERT   [0] nSelectIn (pin 17) --> ASSERT [0] nFault (pin 15)\n"
    "[F6]  READ signal lines\n"
    "[Esc] Exits this utility\n\n");

        while(!kbhit())         /* wait for a key */
        { }
        kp = getkey();          /* get key press */

        switch (kp)             /* select option */
        {
        /* ------------------------------------------------ */
        case (K_F1):            /* stand down all four signals */
            dcr = inportb(REGdcr);
            dcr = (dcr & MASKdcr)
                  | SETBautoFd;
            outportb(REGdcr, dcr);
            break;
        /* ------------------------------------------------ */
        case (K_F2):            /* assert nStrobe */
            dcr = inportb(REGdcr);
            dcr = (dcr & MASKstrobe) | SETBstrobe;
            outportb(REGdcr, dcr);
            break;
        /* ------------------------------------------------ */
        case (K_F3):            /* deassert nAutoFd */
            dcr = inportb(REGdcr);
            dcr = (dcr & MASKautoFd) | 0;
            outportb(REGdcr, dcr);
            break;
        /* ------------------------------------------------ */
        case (K_F4):            /* deassert nInit */
            dcr = inportb(REGdcr);
            dcr = (dcr & MASKnInit) | SETBnInit;
            outportb(REGdcr, dcr);
            break;
        /* ------------------------------------------------ */
        case (K_F5):            /* assert nSelectIn */
            dcr = inportb(REGdcr);
            dcr = (dcr & MASKselectIn) | SETBselectIn;
            outportb(REGdcr, dcr);
            break;
        /* ------------------------------------------------ */
        case (K_F6):            /* read signals */
            dsr = ECPstatus();
            printf("----------------------------"
                   "----------------------------\n");
            if ((dsr & STATnAck) == 0)
            {
                printf("-- [nAck]   is ASSERTED   [0]\n");
            }
            else
            {
                printf("-- [nAck]   is DEASSERTED [1]\n");
            }
            if ((dsr & STATnBusy) == 0)
            {
                printf("-- [busy]   is ASSERTED   [1]\n");
            }
            else
            {
                printf("-- [busy]   is DEASSERTED [0]\n");
            }
            if ((dsr & STATPError) == STATPError)
            {
                printf("-- [PError] is ASSERTED   [1]\n");
            }
            else
            {
                printf("-- [PError] is DEASSERTED [0]\n");
            }
            if ((dsr & STATnFault) == 0)
            {
                printf("-- [nFault] is ASSERTED   [0]\n\n");
            }
            else
            {
                printf("-- [nFault] is DEASSERTED [1]\n\n");
            }
            printf("----------------------------"
                   "----------------------------\n");
            break;
        /* ------------------------------------------------ */
        case (K_Escape):        /* escape from this utility */
            ECPinit();
            return;
        }
    }
}

/*-------------------------------------------------------------------
 * main test program
 *-------------------------------------------------------------------
 */

int main()
{
    int outcome, kp;
    int i, ecr;
    
    printf("\n"
           "----------------------------------------------------------\n"
           " Pseudo-ECP library test program  Copyright (c) 2002,2003\n"
           " K.H. Man <mkh@pl.jaring.my> under terms of GPL Version 2\n"
           "----------------------------------------------------------\n"
           "\n\n");

    /* ----------------------------------------------------
     * test routine perform
     */
    outcome = ECPtest();
    if (outcome != ECP_SUCCESS)         /* reports fatal problems */
    {
        if (outcome & ECP_ERR_NOT_FOUND)
        {
            printf("Fatal: ECP port not found\n");
            return 0;
        }
        if (outcome & ECP_ERR_DIRECTION)
        {
            printf("Fatal: Can't set data direction flag\n");
            return 0;
        }
        if (outcome & ECP_ERR_PWORD8)
        {
            printf("Fatal: Can't handle non 8-bit LPT ports\n");
            return 0;
        }
        if (outcome & ECP_ERR_FIFO)
        {
            printf("Fatal: FIFO of only %d bytes found\n", FIFO);
            return 0;
        }
    }
    /* ----------------------------------------------------
     * report configuration
     */
    printf("ECP port found at [0x%X]\n", REGbase, ecr);

    switch (pwordsz)
    {
        case 0: printf("PWord size is 16 bits\n"); break;
        case 1: printf("PWord size is 8 bits\n"); break;
        case 2: printf("PWord size is 32 bits\n"); break;
    }
    if (outcome & ECP_ERR_CONFLICT)     /* report warnings */
    {
        printf("Warning: Interrupt conflict reported in register\n");
    }
    if (outcome & ECP_ERR_IRQ)
    {
        printf("Warning: IRQ jumper-set, defaults to %d\n", irq);
    }
    else
    {
        printf("IRQ Interrupt is %d\n", irq);
    }
    if (outcome & ECP_ERR_DMA)
    {
        printf("Warning: DMA jumper-set, defaults to %d\n", dma);
    }
    else
    {
        printf("DMA channel is %d\n", dma);
    }
    printf("ECP port FIFO size is %d\n", FIFO);
    printf("\n"
           "---------------------------------------------------------"
           "\n");

    /* ----------------------------------------------------
     * signal and transfer test
     */
    ECPinit();                  /* initialize ECP port, and */
                                /* start listening for incoming */

    while (1)                   /* main loop */
    {
        printf("[Esc] to exit\n"
    "[T] to transmit 300 4000B blocks to a similar remote program\n"
    "[D] to process the same information without transmitting\n"
    "[S] to test the connection of handshake signals\n"
               "\n");

        while(!kbhit())         /* wait for a key */
        {
            if (ECPincoming())  /* try to receive a block */
            {
                receive_test(); /* for now... */
            }
        }
        kp = getkey();          /* get key press */
        
        if (kp == K_Escape)     /* end the program */
        {
            printf("...ECP library test terminated.\n");
            return 0;
        }
        else if ((kp == 'T') || (kp == 't'))
        {                       /* transmission test */
            transmit_test();
        }
        else if ((kp == 'D') || (kp == 'd'))
        {                       /* transmission test */
            dummy_test();
        }
        else if ((kp == 'S') || (kp == 's'))
        {                       /* signal test */
            signal_test();
        }
    }
    ECPexit();                  /* end of test */
}

#endif

/*-------------------------------------------------------------------
 * end of code
 */
