/*-------------------------------------------------------------------
 * nfpt: no-frills parallel transfer
 * a quick-and-dirty pseudo-ECP mode parallel cable transfer utility
 * Copyright (c) 2003 K.H. Man <mkh@pl.jaring.my>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * nfpt header file
 * Last revised: 2003-07-15
 *-------------------------------------------------------------------
 */

#ifdef NFPT_C
#include <stdio.h>
#include <stdlib.h>
#include <pc.h>
#include <keys.h>
#include <strings.h>
#include <unistd.h>
#include <sys/stat.h>
#include <utime.h>
#include <dir.h>
#include "ecp.h"

/*-------------------------------------------------------------------*/

#define NFPT_VER "0.9a"

extern int ECPaddrs[];  /* lookup tables */
extern int IRQlookup[];
extern int DMAlookup[];

/*-------------------------------------------------------------------*/

extern int REGbase;     /* LPT base address, LPT port config info */
extern int irq;         /* IRQ used */
extern int dma;         /* DMA used */
extern int pwordsz;     /* pword size */
extern int FIFO;        /* FIFO size (a slice unit) */

extern int pktcount;    /* packet counter (packet id sequencing) */

/*-------------------------------------------------------------------*/

extern int path_items;          /* number of display items */
extern struct ITEM *path_first; /* display list start pointer */
extern struct ITEM *path_last;  /* display list end pointer */

#endif

/*-------------------------------------------------------------------*/

typedef enum {
    UNCONFIGURED = 0, UNCONNECTED, CONNECTED, READY
} NFPT_MODE;

NFPT_MODE app_state = UNCONFIGURED;

/*-------------------------------------------------------------------*/

struct TAGGEDFILE {		/* findfirst & stat compatible struct */
    char st_name[MAXPATH];	/* file name */
    time_t st_ctime;		/* creation time */
    mode_t st_mode;		/* file mode */
    time_t st_mtime;		/* write time */
    off_t st_size;		/* file size */
    char dummy[12];		/* dummy for packing to FIFO size */
};

#define NFPT_TIMEOUT    4               /* minimum ticks to timeout */
#define NFPT_RETRIES    9               /* maximum number of retries */

#define TEST_DURATION	18		/* test duration in ticks */
#define TEST_PKT_SIZE	MAX_PACKET	/* packet size for testing */

/* struct ITEM **path_sorted = 0;  /* pointer to sorted list */
int tested_rate = 0;            /* tested transfer rate */

/*-------------------------------------------------------------------*/

void port_init(void);
void rate_test(void);

/*-------------------------------------------------------------------*/

int TXdata(struct HEADER *hd, BYTE *buf, int buflen);
int TXfile(struct TAGGEDFILE *tf);

int RXdata(struct HEADER *hd, BYTE *buf, int *buflen);
int RXfile(struct HEADER *rx0, struct TAGGEDFILE *tf);

/*-------------------------------------------------------------------*/

void send_items(void);
void receive_items(void);
void process_incoming(void);
int process_outgoing(char *fname);

/*-------------------------------------------------------------------*/
